/*
 * ********************************************************************************************************************
 *
 *               iFHS7.
 *              ;BBMBMBMc                  rZMBMBR              BMB
 *              MBEr:;PBM,               7MBMMEOBB:             BBB                       RBW
 *     XK:      BO     SB.     :SZ       MBM.       c;;     ir  BBM :FFr       :SSF:    ;xBMB:r   iuGXv.    i:. iF2;
 *     DBBM0r.  :D     S7   ;XMBMB       GMBMu.     MBM:   BMB  MBMBBBMBMS   WMBMBMBBK  MBMBMBM  BMBRBMBW  .MBMBMBMBB
 *      :JMRMMD  ..    ,  1MMRM1;         ;MBMBBR:   MBM  ;MB:  BMB:   MBM. RMBr   sBMH   BM0         UMB,  BMB.  KMBv
 *     ;.   XOW  B1; :uM: 1RE,   i           .2BMBs  rMB. MBO   MBO    JMB; MBB     MBM   BBS    7MBMBOBM:  MBW   :BMc
 *     OBRJ.SEE  MRDOWOR, 3DE:7OBM       .     ;BMB   RMR7BM    BMB    MBB. BMB    ,BMR  .BBZ   MMB   rMB,  BMM   rMB7
 *     :FBRO0D0  RKXSXPR. JOKOOMPi       BMBSSWBMB;    BMBB:    MBMB0ZMBMS  .BMBOXRBMB    MBMDE RBM2;SMBM;  MBB   xBM2
 *         iZGE  O0SHSPO. uGZ7.          sBMBMBDL      :BMO     OZu:BMBK,     rRBMB0;     ,EBMB  xBMBr:ER.  RDU   :OO;
 *     ,BZ, 1D0  RPSFHXR. xWZ .SMr                  . .BBB
 *      :0BMRDG  RESSSKR. 2WOMBW;                   BMBMR
 *         i0BM: SWKHKGO  MBDv
 *           .UB  OOGDM. MK,                                          Copyright (c) 2015-2019.  斯伯坦机器人
 *              ,  XMW  ..
 *                  r                                                                     All rights reserved.
 *
 * ********************************************************************************************************************
 */

package com.sybotan.database

import com.sybotan.base.exceptions.SFieldNotFoundException
import com.sybotan.base.extensions.*
import com.sybotan.database.extensions.filterListByKeys
import com.sybotan.database.parser.SSyntaxParser
import java.io.ByteArrayInputStream
import java.lang.reflect.Field
import java.lang.reflect.ParameterizedType
import javax.persistence.Column
import javax.persistence.Table

/**
 * 数据访问对象基类
 *
 * @author  庞利祥 <sybotan@126.com>
 */
abstract class SAbstractDao<ENTITY: Any>(entityClazz: Class<ENTITY>) {
    companion object {
        /**
         * 获得表名，取实体类@Table注解的名称，如果没有注解，则取实体类类名。
         *
         * @param   clz         mapper类类型
         * @param   keyEsc      转义符，用于避免与数据库关键词冲突
         * @return  表名
         */
        fun tableName(clz: Class<*>, keyEsc: String = ""): String {
            val tableAnno = clz.getAnnotation(javax.persistence.Table::class.java)
            // 如果有@Table注解，则取注解的表名；否则使用类名做为表名
            val tableName = tableAnno?.name ?: clz.simpleName
            val name = tableName.replace(".", "$keyEsc.$keyEsc")
            return "$keyEsc$name$keyEsc"
        } // Fun tableName()

        /**
         * 获得实体类属性对应的数据表字段名
         *
         * @param   clazz   实体类类型
         * @param   name    属性名
         * @return  对应的数据表字段名
         */
        fun columnName(clazz: Class<*>, name: String, keyEsc: String = ""): String {
            try {
                val field = clazz.getAllDeclaredField(name, true, true)
                val colName = field!!.getAnnotation(Column::class.java).name
                return "$keyEsc$colName$keyEsc"
            } catch (e: Exception) {
                throw SFieldNotFoundException(name)
            }
        } // Fun columnName()

        /**
         * 获得实体类属性名
         *
         * @param   clazz   实体类类型
         * @param   name    名称
         * @return  实体类属性名
         */
        fun propertyName(clazz: Class<*>, name: String): String {
            try {
                val field = clazz.getAllDeclaredField(name, true, true)
                return field!!.name
            } catch (e: Exception) {
                throw SFieldNotFoundException(name)
            }
        } // Fun propertyName()
    } // companion object

    /** 实体类类型 */
    val entityClass = entityClazz

    /**
     * 在数据库中插入实体
     *
     * @param   entity      插入数据库的实体
     * @return  插入是否成功
     */
    abstract fun insert(entity: ENTITY): Boolean

    /**
     * 在数据库中插入/更新实体
     *
     * @param   entity      插入/更新数据库的实体
     * @return  插入是否成功
     */
    abstract fun replace(entity: ENTITY): Boolean

    /**
     * 根据条件删除
     *
     * @param   args        删除条件
     * @return  删除是否成功
     */
    abstract fun delete(args: List<SFilter>): Boolean

    /**
     * 删除所有记录
     *
     * @return  删除是否成功
     */
    abstract fun deleteAll(): Boolean

    /**
     * 根据条件更新
     *
     * @param   entity      更新的值
     * @param   projection  将被更新的字段列表（以逗号分隔），为空表示更新所有非空字段
     * @param   args        更新条件
     * @return  更新是否成功
     */
    abstract fun update(entity: ENTITY, projection: ArrayList<String>? = null, args: List<SFilter>? = null): Boolean

    /**
     * 根据条件查询记录
     *
     * @param   args        查询条件
     * @return  查询构造器
     */
    abstract fun select(args: List<SFilter>): SQueryBuilder<ENTITY>

    /**
     * 根据条件查询所有记录
     *
     * @return  查找构建器
     */
    abstract fun selectAll(): SQueryBuilder<ENTITY>

    /**
     * 执行查询操作
     *
     * @param   builder     查询查询构造器
     * @return  查询到的记录
     */
    abstract fun execQuery(builder: SQueryBuilder<ENTITY>): ArrayList<ENTITY>

    /**
     * 执行统计个数操作
     *
     * @param   builder     查询查询构造器
     * @return  查询到的记录数
     */
    abstract fun execCount(builder: SQueryBuilder<ENTITY>): Long

    /**
     * 执行分页查询操作
     *
     * @param   builder     查询查询构造器
     * @param   pageNumber      第几页
     * @param   pageSize        每页记录数
     * @return  查询到的记录
     */
    fun pageQuery(builder: SQueryBuilder<ENTITY>, pageNumber: Int, pageSize: Int): SPageQueryInfo<ENTITY> {
        val pageInfo = SPageQueryInfo<ENTITY>(pageNumber, pageSize)
        builder.limit((pageNumber - 1) * pageSize, pageSize)
        pageInfo.total = execCount(builder)
        pageInfo.content = execQuery(builder)

        return pageInfo
    } // Fun pageQuery()

    /**
     * 删除实体
     *
     * @param   entity      被删除的实体
     * @return  删除是否成功
     */
    open fun delete(entity: ENTITY): Boolean {
        return delete(entity.filterListByKeys())
    } // Fun delete()

    /**
     * 根据条件删除
     *
     * @param   args        删除条件
     * @return  删除是否成功
     */
    open fun delete(vararg args: Pair<String, Any>): Boolean {
        val filterList = ArrayList<SFilter>()
        for (arg in args) {     // 同时取列表元素与下标
            filterList.add(SFilter.eq(arg.first, arg.second))
        }

        return delete(filterList)
    } // Fun delete()

    /**
     * 根据条件删除
     *
     * @param   args        删除条件
     * @return  删除是否成功
     */
    open fun delete(vararg args: SFilter): Boolean {
        return delete(args.toList())
    } // Fun delete()

    /**
     * 根据条件删除
     *
     * @param   query       删除条件
     * @return  删除是否成功
     */
    open fun delete(query: String): Boolean {
        val parser = SSyntaxParser(ByteArrayInputStream(query.toByteArray()))
        return delete(parser.parserFilters())
    } // Fun delete()

    /**
     * 更新实体
     *
     * @param   entity      更新实体
     * @param   projection  将被更新的字段列表（以逗号分隔），为空表示更新所有非空字段
     * @return  更新是否成功
     */
    open fun update(entity: ENTITY, projection: ArrayList<String>? = null): Boolean {
        return update(entity, projection, entity.filterListByKeys())
    } // Fun update()

    /**
     * 根据条件更新
     *
     * @param   entity      更新实体内容
     * @param   projection  将被更新的字段列表（以逗号分隔），为空表示更新所有非空字段
     * @param   args        更新条件
     * @return  更新是否成功
     */
    open fun update(entity: ENTITY, projection: ArrayList<String>? = null, vararg args: Pair<String, Any>): Boolean {
        val filterList = ArrayList<SFilter>()
        for (arg in args) {
            filterList.add(SFilter.eq(arg.first, arg.second))
        }

        return update(entity, projection, filterList)
    } // Fun update()

    /**
     * 根据条件更新
     *
     * @param   entity      更新实体内容
     * @param   projection  将被更新的字段列表（以逗号分隔），为空表示更新所有非空字段
     * @param   args        更新条件
     * @return  更新是否成功
     */
    open fun update(entity: ENTITY, projection: ArrayList<String>? = null, vararg args: SFilter): Boolean {
        return update(entity, projection, args.toList())
    } // Fun update()

    /**
     * 根据条件更新
     *
     * @param   entity      更新实体内容
     * @param   projection  将被更新的字段列表（以逗号分隔），为空表示更新所有非空字段
     * @param   query       更新条件
     * @return  更新是否成功
     */
    open fun update(entity: ENTITY, projection: ArrayList<String>? = null, query: String): Boolean {
        val parser = SSyntaxParser(ByteArrayInputStream(query.toByteArray()))
        return update(entity, projection, parser.parserFilters())
    } // Fun update()

    /**
     * 根据条件查询一条记录
     *
     * @param   args        查询条件
     * @return  查询构造器
     */
    open fun select(vararg args: Pair<String, Any>): SQueryBuilder<ENTITY> {
        val filterList = ArrayList<SFilter>()
        for (arg in args) {     // 同时取列表元素与下标
            filterList.add(SFilter.eq(arg.first, arg.second))
        }

        return select(filterList)
    } // Fun select()

    /**
     * 根据条件查询记录
     *
     * @param   args        查询条件
     * @return  查询构造器
     */
    open fun select(vararg args: SFilter): SQueryBuilder<ENTITY> {
        return select(args.toList())
    } // Fun select()

    /**
     * 根据条件查询记录
     *
     * @param   query        查询条件
     * @return  查询构造器
     */
    open fun select(query: String?): SQueryBuilder<ENTITY> {
        // 如果查询条件为空，则执行查询所有
        if (query == null) {
            return selectAll()
        }

        val queryParser = SSyntaxParser(ByteArrayInputStream(query.toByteArray()))
        val filters  = queryParser.parserFilters()
        return select(filters)
    } // Fun select()

    /**
     * 获得表名
     */
    fun tableName(): String {
        val clz = (javaClass.genericSuperclass as ParameterizedType).actualTypeArguments[0] as Class<*>
        if (clz.isAnnotationPresent(Table::class.java)) {
            val anno = clz.getAnnotation(Table::class.java)
            return anno!!.name
        }

        return clz.simpleName
    } // Fun tableName()

    /**
     * 实体的field转换为Pair结构
     *
     * @param   entity      要转换的实体对象
     * @param   field       实体对应的域
     */
    protected fun fieldToPair(entity: ENTITY, field: Field):  Pair<String, Any?> {
        // 获得field对应的数据库表字段名
        val anno = field.getAnnotation(Column::class.java)
        return Pair<String, Any?>(anno.name, when (field.type.simpleName) {
            "byte", "Byte" -> field.getByte(entity)
            "double", "Double" -> field.getDouble(entity)
            "float", "Float" -> field.getFloat(entity)
            "int", "Integer", "Int" -> field.getInt(entity)
            "long", "Long" -> field.getLong(entity)
            "short", "Short" -> field.getShort(entity)
            "Boolean" -> field.getBoolean(entity)
            "String" ->  field.get(entity)
            else ->{
                if (entity == null){
                    null
                }else{
                    if (field.get(entity) == null){
                        null
                    }else{
                        field.get(entity).toJson()
                    }
                }
            }
        })
    } // Fun fieldToPair()

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // 级联操作相关
    /**
     * 查询级联关系
     *
     * @param   table       表名
     * @param   id1         对象表ID字段名
     * @param   id2         级联对象ID字段名
     * @param   self        自身级联
     * @param   filter      附加过滤条件
     * @param   idList      对象ID列表
     * @return  关系列表
     */
    abstract fun relationQuery(table: String, id1: String, id2: String, self: Boolean, filter: String, idList: ArrayList<Any>): ArrayList<HashMap<String, Any>>
} // Class SAbstractDao
